// laser_mouse.h

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <fcntl.h>
#include <getopt.h>
#include <png.h>
#include <jpeglib.h>
#include <sys/ioctl.h>
#include <linux/videodev.h>
#include <SDL/SDL.h>
#include <sys/time.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#define XK_LATIN1
#define XK_MISCELLANY
#define XK_XKB_KEYS
#include <X11/keysymdef.h>

char prog[] = PROG;
char program[] = PROGRAM;
char version[] = VERSION;

#ifndef laser_mouse_h
#define laser_mouse_h

enum Save_Types {
  PNG = 0,
  PPM = 1,
  JPEG = 2
};

enum Ppm_Types {
  RAW = 0,
  ASCII = 1
};

struct Options
{
  int optionfile;
  char file[256];

  char video_device[256];
  int video_channel;

  int video_brightness;
  int video_color;
  int video_contrast;
  int video_hue;
  int video_whiteness;

  int filter_autobright;
  int filter_polarize,filter_polarize_show;
  int filter_swaprgb;

  int savetype;
  int savetype_jpg_optimize;
  int savetype_jpg_quality;
  int savetype_jpg_smoothness;
  int savetype_png_compression;
  png_byte savetype_png_interlace;
  int savetype_ppm_format;

  int capturexy;
  int capturex0,capturey0,capturex1,capturey1;

  int vid_winxy;
  int vid_winx0,vid_winy0,vid_winx1,vid_winy1;

  int mousexy;
  int mousex0,mousey0,mousex1,mousey1;

  int mousedown_starttime_max;
};

struct SDL_Screen
{
  int width,height,size,depth,sdl_flags;
  SDL_Surface *display;
  Uint8 *pixelbuf;
  SDL_Rect rect;
};

struct VideoPaint
{
  struct Options option;

// Paint
  struct SDL_Screen screen;
  int old_x, old_y, mouse_x, mouse_y, mouse_button;
  int mousedown_start;
  struct timeval mousedown_starttime;

// Video
  struct SDL_Screen video;
  int video_show;
  int dev;
  int dump;
  int active;
  int mousedown;
  unsigned char *pic;
  unsigned char *pic_temp;
  char savefile[256];

  struct video_capability vid_caps;
//char name[32];
//int type;
//int channels;	/* Num channels */
//int audios;	/* Num audio devices */
//int maxwidth;	/* Supported width */
//int maxheight;	/* And height */
//int minwidth;	/* Supported width */
//int minheight;	/* And height */
  struct video_channel vid_chan;
//int channel;
//char name[32];
//int tuners;
//__u32  flags;
//#define VIDEO_VC_TUNER		1	/* Channel has a tuner */
//#define VIDEO_VC_AUDIO		2	/* Channel has audio */
//__u16  type;
//#define VIDEO_TYPE_TV		1
//#define VIDEO_TYPE_CAMERA	2
//__u16 norm;			/* Norm set by channel */
  struct video_window vid_win;
//__u32	x,y;			/* Position of window */
//__u32	width,height;		/* Its size */
//__u32	chromakey;
//__u32	flags;
//struct	video_clip __user *clips;	/* Set only */
//int	clipcount;
//#define VIDEO_WINDOW_INTERLACE	1
//#define VIDEO_WINDOW_CHROMAKEY	16	/* Overlay by chromakey */
//#define VIDEO_CLIP_BITMAP	-1
// /* bitmap is 1024x625, a '1' bit represents a clipped pixel */
//#define VIDEO_CLIPMAP_SIZE	(128 * 625)
  struct video_picture vid_pic;
//__u16	brightness;
//__u16	hue;
//__u16	colour;
//__u16	contrast;
//__u16	whiteness;	/* Black and white only */
//__u16	depth;		/* Capture depth */
//__u16   palette;	/* Palette in use */
//#define VIDEO_PALETTE_GREY	1	/* Linear greyscale */
//#define VIDEO_PALETTE_HI240	2	/* High 240 cube (BT848) */
//#define VIDEO_PALETTE_RGB565	3	/* 565 16 bit RGB */
//#define VIDEO_PALETTE_RGB24	4	/* 24bit RGB */
//#define VIDEO_PALETTE_RGB32	5	/* 32bit RGB */
//#define VIDEO_PALETTE_RGB555	6	/* 555 15bit RGB */
//#define VIDEO_PALETTE_YUV422	7	/* YUV422 capture */
//#define VIDEO_PALETTE_YUYV	8
//#define VIDEO_PALETTE_UYVY	9	/* The great thing about standards is ... */
//#define VIDEO_PALETTE_YUV420	10
//#define VIDEO_PALETTE_YUV411	11	/* YUV411 capture */
//#define VIDEO_PALETTE_RAW	12	/* RAW capture (BT848) */
//#define VIDEO_PALETTE_YUV422P	13	/* YUV 4:2:2 Planar */
//#define VIDEO_PALETTE_YUV411P	14	/* YUV 4:1:1 Planar */
//#define VIDEO_PALETTE_YUV420P	15	/* YUV 4:2:0 Planar */
//#define VIDEO_PALETTE_YUV410P	16	/* YUV 4:1:0 Planar */
//#define VIDEO_PALETTE_PLANAR	13	/* start of planar entries */
//#define VIDEO_PALETTE_COMPONENT 7	/* start of component entries */
//  struct video_clip vid_clips[32];
//__s32	x,y;
//__s32	width, height;
//struct	video_clip *next;	/* For user use/driver use only */
};

#define min(a,b) (a < b ? a : b)

#define max(a,b) (a > b ? a : b)

//  b = 1.164 * (y - 16)                         + 2.018 * (cb - 128);
//  g = 1.164 * (y - 16)  -  0.813 * (cr - 128)  - 0.391 * (cb - 128);
//  r = 1.164 * (y - 16)  +  1.596 * (cr - 128);
//  b=(2384*(y-16)+4133*(cb-128))/2048
//  g=(2384*(y-16)-1665*(cr-128)+801*(cb-128))/2048
//  r=(2384*(y-16)+3269*(cr-128))/2048
#define YUV4222RGB(y0, cr, cb, r, g, b, y1)\
  y1 = ((y0 << 11) + (y0 << 8) + (y0 << 6) + (y0 << 4)) >> 11;\
  r = y1 + ( ((cb << 12) + (cb <<  5) + (cb << 2) + cb) >> 11);\
  g = y1 - ( ((cr << 10) + (cr <<  9) + (cr << 7) + cr) >> 11 ) \
         + ( ((cb <<  9) + (cb <<  8) + (cb << 5) + cb) >> 11 );\
  b = y1 + ( ((cr << 11) + (cr << 10) + (cr << 7) + (cr << 6) + (cr << 2) + cr) >> 11);\
  r = r < 0 ? 0 : r;\
  g = g < 0 ? 0 : g;\
  b = b < 0 ? 0 : b;\
  r = r > 255 ? 255 : r;\
  g = g > 255 ? 255 : g;\
  b = b > 255 ? 255 : b

/*macro used to convert a YUV pixel to RGB format
   from Bart Nabbe
   v*1434 / 2048
   (u*406 - v*595)/2048
   u*2078 / 2048
   bitshift performance enhanced by Damien Douxchamps
*/

#define YUV2RGB(y, u, v, r, g, b)\
  r = y + ( ((v << 10) + (v << 8) + (v << 7) + (v << 4) + (v << 3) + (v << 1)) >> 11);\
  g = y - ( ((u << 8) + (u << 7) + (u << 4) + (u << 2) + (u << 1)) >> 11 ) - \
          ( ((v << 9) + (v << 6) + (v << 4) + (v << 1) + v) >> 11 );\
  b = y + ( ((u << 11) + (u << 5) - (u << 1)) >> 11);\
  r = r < 0 ? 0 : r;\
  g = g < 0 ? 0 : g;\
  b = b < 0 ? 0 : b;\
  r = r > 255 ? 255 : r;\
  g = g > 255 ? 255 : g;\
  b = b > 255 ? 255 : b

#define RGB2YUV(r, g, b, y, u, v)\
  y = (9798*r + 19235*g + 3736*b)  >> 15;\
  u = ((-4784*r - 9437*g + 14221*b) >> 15)  + 128;\
  v = ((20218*r - 16941*g - 3277*b) >> 15) + 128;\
  y = y < 0 ? 0 : y;\
  u = u < 0 ? 0 : u;\
  v = v < 0 ? 0 : v;\
  y = y > 255 ? 255 : y;\
  u = u > 255 ? 255 : u;\
  v = v > 255 ? 255 : v

void MOUSEMOTION(struct VideoPaint *vp, int x, int y);
void MOUSEBUTTONUP(struct VideoPaint *vp);
void MOUSEBUTTONDOWN(struct VideoPaint *vp);

#endif
