// laser_mouse_video.c

/* ************************************************************** */
void video_open_cam(struct VideoPaint *vp)
{
  if (vp->dev<=0)
  {
    vp->dev = open(vp->option.video_device,O_RDWR);
    if (vp->dev < 0)
    {
      perror(vp->option.video_device);
      exit(1);
    }
  }
}

/* ************************************************************** */
void video_close_cam(struct VideoPaint *vp,int quiting)
{
  int debug = 0;
  if (vp->dev > 0)
  {
    close(vp->dev);
    vp->dev = 0;
  }
}

/* ************************************************************** */
void video_get_cam_info(struct VideoPaint *vp, int first)
{
  ioctl(vp->dev,VIDIOCGCAP,&vp->vid_caps);
  ioctl(vp->dev,VIDIOCGWIN,&vp->vid_win);
  ioctl(vp->dev,VIDIOCGPICT,&vp->vid_pic);
  if (vp->vid_caps.channels > 1)
    ioctl(vp->dev,VIDIOCGCHAN,&vp->vid_chan);

if (first)
{
  int i;
  struct video_clip vid_clips[32];

  for (i = 0; i < 32; i++)
  {
    vid_clips[i].x      = 0;
    vid_clips[i].y      = 0;
    vid_clips[i].width  = 0;
    vid_clips[i].height = 0;
  }
  vp->vid_win.clips = vid_clips;
  vp->vid_win.clipcount = 0;

  if (vp->vid_caps.type & VID_TYPE_MONOCHROME)
  {
    fprintf(stderr,"Error: Monochrome device");
    exit(-1);
  }
  else
  {
    if (vp->vid_pic.palette == VIDEO_PALETTE_RGB24) /* 4-RGB24 QUICKCAM */
    {
      vp->pic = (unsigned char*)
        realloc(vp->pic,vp->vid_caps.maxwidth*vp->vid_caps.maxheight*3);
      vp->pic_temp = (unsigned char*)
        realloc(vp->pic_temp,vp->vid_caps.maxwidth*vp->vid_caps.maxheight*3);
    }
    else if (vp->vid_pic.palette == VIDEO_PALETTE_UYVY) /* 9-UYVY HAUPAUGE B878 */
    {
      vp->pic = (unsigned char*)
        realloc(vp->pic,vp->vid_caps.maxwidth*vp->vid_caps.maxheight*3);
      vp->pic_temp = (unsigned char*)
        realloc(vp->pic_temp,vp->vid_caps.maxwidth*vp->vid_caps.maxheight*3);
    }
    else if (vp->vid_pic.palette == VIDEO_PALETTE_YUV422) /* 7-YUV422 HAUPAUGE B878 */
    {
      vp->pic = (unsigned char*)
        realloc(vp->pic,vp->vid_caps.maxwidth*vp->vid_caps.maxheight*3);
      vp->pic_temp = (unsigned char*)
        realloc(vp->pic_temp,vp->vid_caps.maxwidth*vp->vid_caps.maxheight*3);
    }
    else
    {
      fprintf(stderr,
        "Error: vid_pic.palette should 4-RGB24 7-YUV422 9-UYVY, but is %d !",vp->vid_pic.palette);
    }
  }
}
}

int video_set_cam_vid_pic(struct VideoPaint *vp)
{
  return ioctl(vp->dev,VIDIOCSPICT,&vp->vid_pic);
}

int video_set_cam_vid_chan(struct VideoPaint *vp)
{
  return ioctl(vp->dev,VIDIOCSCHAN,&vp->vid_chan);
}

int video_set_cam_vid_win(struct VideoPaint *vp)
{
  return ioctl(vp->dev,VIDIOCSWIN,&vp->vid_win);
}

/* ************************************************************** */
void video_set_cam_info(struct VideoPaint *vp)
{
  if (video_set_cam_vid_pic(vp) < 0)    perror ("ioctl (VIDIOCSPICT)");
  if (vp->vid_caps.channels > 1)
    if (video_set_cam_vid_chan(vp) < 0) perror ("ioctl (VIDIOCSCHAN)");
  if (video_set_cam_vid_win(vp) < 0)    perror ("ioctl (VIDIOCSWIN)");
}

/* ************************************************************** */
void video_print_cam_info(struct VideoPaint *vp)
{
  fprintf(stdout,"Name: %s\n",vp->vid_caps.name);
  fprintf(stdout,"Type: %i\n",vp->vid_caps.type);
  if (vp->vid_caps.type & VID_TYPE_CAPTURE)
    fprintf(stdout,"\tCan capture\n");
  if (vp->vid_caps.type & VID_TYPE_TUNER)
    fprintf(stdout,"\tCan tune\n");
  if (vp->vid_caps.type & VID_TYPE_TELETEXT)
    fprintf(stdout,"\tDoes teletext\n");
  if (vp->vid_caps.type & VID_TYPE_OVERLAY)
    fprintf(stdout,"\tOverlay onto frame buffer\n");
  if (vp->vid_caps.type & VID_TYPE_CHROMAKEY)
    fprintf(stdout,"\tOverlay by chromakey\n");
  if (vp->vid_caps.type & VID_TYPE_CLIPPING)
    fprintf(stdout,"\tCan clip\n");
  if (vp->vid_caps.type & VID_TYPE_FRAMERAM)
    fprintf(stdout,"\tUses the frame buffer memory\n");
  if (vp->vid_caps.type & VID_TYPE_SCALES)
    fprintf(stdout,"\tScalable\n");
  if (vp->vid_caps.type & VID_TYPE_MONOCHROME)
    fprintf(stdout,"\tMonochrome only\n");
  if (vp->vid_caps.type & VID_TYPE_SUBCAPTURE)
    fprintf(stdout,"\tCan capture subareas of the image\n");
  fprintf(stdout,"Channels: %i\n",vp->vid_caps.channels);
  fprintf(stdout,"Audios: %i\n",vp->vid_caps.audios);
  fprintf(stdout,"Maxwidth: %i\n",vp->vid_caps.maxwidth);
  fprintf(stdout,"Maxheight: %i\n",vp->vid_caps.maxheight);
  fprintf(stdout,"Minwidth: %i\n",vp->vid_caps.minwidth);
  fprintf(stdout,"Minheight: %i\n",vp->vid_caps.minheight);
  if (vp->vid_caps.channels > 1)
  {
    fprintf(stdout,"---------\n");
    fprintf(stdout,"Channel: %i\n",vp->vid_chan.channel);
    fprintf(stdout,"Name: %s\n",vp->vid_chan.name);
  }
  fprintf(stdout,"---------\n");
  fprintf(stdout,"X: %i\n",vp->vid_win.x);
  fprintf(stdout,"Y: %i\n",vp->vid_win.y);
  fprintf(stdout,"Width: %i\n",vp->vid_win.width);
  fprintf(stdout,"Height: %i\n",vp->vid_win.height);
  fprintf(stdout,"Chromakey: %i\n",vp->vid_win.chromakey);
  fprintf(stdout,"Flags: %i\n",vp->vid_win.flags);
  fprintf(stdout,"---------\n");
  fprintf(stdout,"Brightness:\t%i (%i)\n",vp->vid_pic.brightness,vp->vid_pic.brightness/256);
  fprintf(stdout,"Hue:\t\t%i (%i)\n",vp->vid_pic.hue,vp->vid_pic.hue/256);
  fprintf(stdout,"Color:\t\t%i (%i)\n",vp->vid_pic.colour,vp->vid_pic.colour/256);
  fprintf(stdout,"Contrast:\t%i (%i)\n",vp->vid_pic.contrast,vp->vid_pic.contrast/256);
  fprintf(stdout,"Whiteness:\t%i (%i)\n",vp->vid_pic.whiteness,vp->vid_pic.whiteness/256);
  fprintf(stdout,"Depth:\t\t%i\n",vp->vid_pic.depth);
  fprintf(stdout,"Palette:\t%i\n",vp->vid_pic.palette);
}

/* ************************************************************** */
void video_filter_capturexy(struct VideoPaint *vp)
{
  unsigned char *p = vp->pic;
  int size = vp->video.size;
  int i,j,bytes,linesize;

  switch (vp->video.depth)
  {
  case 24 : bytes = 3; break;
  case 16 : bytes = 2; break;
  default : return;
  }

  linesize = vp->video.width * bytes;

  if (vp->option.capturey0 > 0)
    memset((char*) p,0,vp->option.capturey0*linesize);

  for (j=vp->option.capturey0*linesize; j<=vp->option.capturey1*linesize; j+=linesize)
  {
    if (vp->option.capturex0)
      memset((char*)&p[j],0,(vp->option.capturex0-1)*bytes);
    if (vp->option.capturex1 < (vp->video.width-1))
      memset((char*)&p[j+(vp->option.capturex1+1)*bytes],0,(vp->video.width-(vp->option.capturex1+1))*bytes);
  }

  if (vp->option.capturey1 < (vp->video.height-1))
    memset((char*)&p[(vp->option.capturey1+1)*linesize],0,
      (vp->video.height-1-vp->option.capturey1)*linesize);
}

/* ************************************************************** */
/* routine to convert an array of YUV data to RGB format from Bart Nabbe */
inline void video_filter_uyvy2rgb (struct VideoPaint *vp)
{
  unsigned char *p = vp->pic;
  unsigned char *YUV = vp->pic_temp;
  int size = vp->video.size;
  int i, j;
  register int y0, y1, u, v;
  register int r, g, b;

  memcpy((char*)YUV,p,size * 2);

  for (i = 0, j = 0; i < (size << 1); i += 4, j += 6)
  {
    u = (unsigned char) YUV[i + 0] - 128;
    y0 = (unsigned char) YUV[i + 1];
    v = (unsigned char) YUV[i + 2] - 128;
    y1 = (unsigned char) YUV[i + 3];
    YUV2RGB (y0, u, v, r, g, b);
    p[j + 2] = r; // swap
    p[j + 1] = g;
    p[j + 0] = b; // swap
    YUV2RGB (y1, u, v, r, g, b);
    p[j + 5] = r; // swap
    p[j + 4] = g;
    p[j + 3] = b; // swap
  }
}

/* ************************************************************** */
inline void video_filter_yuv4222rgb (struct VideoPaint *vp)
{
  unsigned char *p = vp->pic;
  unsigned char *YUV = vp->pic_temp;
  int size = vp->video.size;
  int i, j;
  register int y, cr, cb, y00, y01, y0;
  register int r, g, b;

  memcpy((char*)YUV,p,size * 2);

  for (i = 0, j = 0; i < (size << 1); i += 2, j += 3)
  {
    cb = (unsigned char) YUV[i + 0] & (unsigned char) 16;
    y00 = (unsigned char) YUV[i + 0] & (unsigned char) 128;
    cr = (unsigned char) YUV[i + 1] & (unsigned char) 16;
    y01 = (unsigned char) YUV[i + 1] & (unsigned char) 128;
    YUV4222RGB(y00+y01, cr, cb, r, g, b, y0);
    p[j + 0] = r;
    p[j + 1] = g;
    p[j + 2] = b;
  }
}

/* ************************************************************** */
void video_filter_autobright(struct VideoPaint *vp)
{
  int i, average;
  unsigned char *p = vp->pic;
  int size = vp->video.size*3;
  long total = 0;

  for (i=0; i<size; i++) total += *p++;

  average = total / size;
  if ((average <= 124) || (average >= 132))
    vp->vid_pic.brightness += ((128 - average)/3) * 256;
}

/* ************************************************************** */
void video_filter_swaprgb(struct VideoPaint *vp)
{
  unsigned char  c;
  unsigned char *p = vp->pic;
  int size = vp->video.size;
  int bytes_per_pixel = 3;

  while (--size)
  {
    c = p[0]; p[0] = p[2]; p[2] = c;
    p += bytes_per_pixel;
  }
}

/* ************************************************************** */
void video_filter_polarize(struct VideoPaint *vp)
{
  unsigned char  c;
  unsigned char *p = vp->pic;
  int i,x,y;
  int size = vp->video.size;
  int bytes_per_pixel = 3;
  int minx = vp->video.width;
  int maxx = -1;
  int miny = vp->video.height;
  int maxy = -1;
  int pix;
  int avgxy = -1;
  int x0,y0,x1,y1;
  int mousex,mousey;
  int pixmin = 2;
  int pixmax = vp->video.width*vp->video.height/16;

  for (i=0, pix=0;
    (i<size) && (vp->option.filter_polarize_show || (pix<pixmax));
    i++,p+=bytes_per_pixel)
    if ((p[0] >= vp->option.filter_polarize)
    || (p[1] >= vp->option.filter_polarize)
    || (p[2] >= vp->option.filter_polarize))
    {
      p[0] = p[1] = p[2] = 255;
      x = i % vp->video.width;
      y = i / vp->video.width;
      minx = min(minx,x);
      maxx = max(maxx,x);
      miny = min(miny,y);
      maxy = max(maxy,y);
      pix++;
    }
    else
      p[0] = p[1] = p[2] = 0;

  if( pix >= pixmin                       // mind. pixmin Punkte
  && ((maxx-minx)<(vp->video.width/4))    // schmaler als 1/4
  && ((maxy-miny)<(vp->video.height/4))   // niedriger als 1/4
  && (pix >= ((maxx-minx)*(maxy-miny)/8)) // im Rechteck mind. 1/8 Punkte
  //&& (pix >= ((maxx-minx)*(maxy-miny)/4)) // im Rechteck mind. 1/4 Punkte
  )
  {
    mousex = (minx + maxx) / 2;
    mousey = (miny + maxy) / 2;
    avgxy = mousey * vp->video.width + mousex;

    x0 = vp->option.capturex0;
    y0 = vp->option.capturey0;
    x1 = vp->option.capturex1;
    y1 = vp->option.capturey1;

    if ((mousex >= x0) && (mousey >= y0) && (mousex <= x1) && (mousey <= y1))
    {
      mousex=vp->option.mousex0
        + ((mousex-x0)*(vp->option.mousex1-vp->option.mousex0))/(x1-x0);
      mousey=vp->option.mousey0
        + ((mousey-y0)*(vp->option.mousey1-vp->option.mousey0))/(y1-y0);
      if (vp->video_show)
        if (x0 && y0 
        && (x1 < (vp->video.width-1))
        && (y1 < (vp->video.height-1)))
        {
          p = vp->pic;
          p += (avgxy-vp->video.width-1) * bytes_per_pixel;
          for (i=0; i<(bytes_per_pixel*3); i++) p[i] = ( i%3 ? 255 : 0 );
          p += vp->video.width * bytes_per_pixel;
          for (i=0; i<(bytes_per_pixel*3); i++) p[i] = ( i%3 ? 255 : 0 );
          p += vp->video.width * bytes_per_pixel;
          for (i=0; i<(bytes_per_pixel*3); i++) p[i] = ( i%3 ? 255 : 0 );
        }
    }
    else avgxy = -1;
  }
  if (avgxy >= 0)
  {
    if ((! vp->mousedown) && (! vp->mousedown_start))
    {
      set_usecs_in(&vp->mousedown_starttime);
      vp->mousedown_start = 1;
    }
    else if ((! vp->mousedown) && vp->mousedown_start)
    {
      int delta = get_usecs_since(&vp->mousedown_starttime);
      if (vp->option.mousedown_starttime_max < delta)
      {
        fprintf(stdout,"%d: mouse move %d %d before down\n",vp->active,mousex,mousey);
        if (vp->active) MOUSEMOTION(vp,mousex,mousey);
        fprintf(stdout,"%d: mouse button down\n",vp->active);
        vp->mousedown = 1;
        if (vp->active) MOUSEBUTTONDOWN(vp);
      }
      else
      {
        fprintf(stdout,"%d: mouse move %d %d before down (%d)\n",vp->active,mousex,mousey,delta);
        if (vp->active) MOUSEMOTION(vp,mousex,mousey);
      }
    }
    else
    {
      fprintf(stdout,"%d: mouse move %d %d\n",vp->active,mousex,mousey);
      if (vp->active) MOUSEMOTION(vp,mousex,mousey);
    }
  }
  else
  {
    vp->mousedown_start = 0;
    if (vp->mousedown)
    {
      fprintf(stdout,"%d: mouse button up\n",vp->active);
      vp->mousedown = 0;
      if (vp->active) MOUSEBUTTONUP(vp);
    }
  }
}

/* ************************************************************** */
void video_png_save(struct VideoPaint *vp)
{
  FILE *fp;
  png_structp png_ptr;
  png_infop info_ptr;
  png_uint_32 k, height, width;
  png_bytep row_pointers[vp->video.height];

  if (!strcmp(vp->savefile,"")) fp = stdout;
  else fp = fopen(vp->savefile,"w");

  if (fp == NULL)
  {
    fprintf(stderr, "Error: Could not open file\n");
    return;
  }

  /* Create and initialize the png_struct with the desired error handler
   * functions.  If you want to use the default stderr and longjump method,
   * you can supply NULL for the last three parameters.  We also check that
   * the library version is compatible with the one used at compile time,
   * in case we are using dynamically linked libraries.  REQUIRED.
   */
  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);

  if (png_ptr == NULL)
  {
    fprintf(stderr,"Error: could not create PNG\n");
    fclose(fp);
    return;
  }

  /* Allocate/initialize the image information data.  REQUIRED */
  info_ptr = png_create_info_struct(png_ptr);
  if (info_ptr == NULL)
  {
    fclose(fp);
    png_destroy_write_struct(&png_ptr,(png_infopp)NULL);
    return;
  }

  /* Set error handling.  REQUIRED if you aren't supplying your own
   * error hadnling functions in the png_create_write_struct() call.
   */
  if (setjmp(png_ptr->jmpbuf))
  {
    /* If we get here, we had a problem reading the file */
    fclose(fp);
    png_destroy_write_struct(&png_ptr,(png_infopp)NULL);
    return;
  }

  /* One of the following I/O initialization functions is REQUIRED */
  /* set up the output control if you are using standard C streams */
  png_init_io(png_ptr,fp);

  /* Set the image information here.  Width and height are up to 2^31,
   * bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
   * the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
   * PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
   * or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
   * PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
   * currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
   */
  png_set_IHDR(png_ptr,info_ptr,vp->video.width,vp->video.height,8,
    PNG_COLOR_TYPE_RGB,vp->option.savetype_png_interlace,
    PNG_COMPRESSION_TYPE_BASE,PNG_FILTER_TYPE_BASE);
  png_set_compression_level(png_ptr,vp->option.savetype_png_compression);

  /* Optionally write comments into the image */
/*
  text_ptr[0].key = "Title";
  text_ptr[0].text = "Mona Lisa";
  text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
  text_ptr[1].key = "Author";
  text_ptr[1].text = "Leonardo DaVinci";
  text_ptr[1].compression = PNG_TEXT_COMPRESSION_NONE;
  text_ptr[2].key = "Description";
  text_ptr[2].text = "<long text>";
  text_ptr[2].compression = PNG_TEXT_COMPRESSION_zTXt;
  png_set_text(png_ptr, info_ptr, text_ptr, 3);
*/
  /* Write the file header information.  REQUIRED */
  png_write_info(png_ptr,info_ptr);
  /* Once we write out the header, the compression type on the text
   * chunks gets changed to PNG_TEXT_COMPRESSION_NONE_WR or
   * PNG_TEXT_COMPRESSION_zTXt_WR, so it doesn't get written out again
   * at the end.
   */
  png_set_bgr(png_ptr);
  /* The easiest way to write the image (you may have a different memory
   * layout, however, so choose what fits your needs best).  You need to
   * use the first method if you aren't handling interlacing yourself.
   */
/*
  png_uint_32 k, height, width;
  png_byte image[height][width];
  png_bytep row_pointers[height];
*/
  for (k = 0; k < vp->video.height; k++)
    row_pointers[k] = vp->pic + k * vp->video.width * 3;
  /* One of the following output methods is REQUIRED */
  png_write_image(png_ptr,row_pointers);
  /* You can write optional chunks like tEXt, zTXt, and tIME at the end
   * as well.
   */
  /* It is REQUIRED to call this to finish writing the rest of the file */
  png_write_end(png_ptr,info_ptr);
  /* if you malloced the palette, free it here */
  free(info_ptr->palette);
  /* if you allocated any text comments, free them here */
  /* clean up after the write, and free any memory allocated */
  png_destroy_write_struct(&png_ptr,(png_infopp)NULL);
  fclose(fp);
  return;
}

/* ************************************************************** */
void video_ppm_save(struct VideoPaint *vp)
{
  FILE *outfile;
  unsigned char buff[3];
  int i;

  if (!strcmp(vp->savefile,"")) outfile = stdout;
  else outfile = fopen(vp->savefile,"w");

  if (outfile == NULL)
  {
    fprintf(stderr,"Error: Could not open file");
    return;
  }

  if (vp->option.savetype_ppm_format == RAW)
  {
    fprintf(outfile,"P6\n%d %d\n%d\n",vp->video.width,vp->video.height,255);
    for (i = 0; i < vp->video.size * 3; i++)
      fputc(vp->pic[i], outfile);
  }
  else if (vp->option.savetype_ppm_format == ASCII)
  {
    fprintf (outfile,"P3\n%d %d\n%d\n",vp->video.width,vp->video.height,255);
    for (i = 0; i < vp->video.size * 3; i=i+3)
      fprintf(outfile,"%03d %03d %03d\n",vp->pic[i],vp->pic[i+1],vp->pic[i+2]);
  }

  fclose(outfile);
}

/* ************************************************************** */
void video_jpeg_save(struct VideoPaint *vp)
{
  FILE *outfile;
  int ls=0, x, y;
  int i;
  int lines, pixels_per_line;
  unsigned char buff[3];
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JSAMPROW row_pointer[1];     // pointer to JSAMPLE row(s)
  int row_stride;              // physical row width in image buffer

  if (! strcmp(vp->savefile,"")) outfile = stdout;
  else outfile = fopen(vp->savefile,"w");

  if (outfile == NULL)
  {
    fprintf(stderr, "Error: Could not open file");
    return;
  }

  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);

  jpeg_stdio_dest(&cinfo,outfile);
  cinfo.image_width = vp->video.width;
  cinfo.image_height = vp->video.height;

  cinfo.input_components = 3;
  cinfo.in_color_space = JCS_RGB;

  cinfo.smoothing_factor = vp->option.savetype_jpg_smoothness;
  if (vp->option.savetype_jpg_smoothness)
    cinfo.optimize_coding = TRUE;
  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo, vp->option.savetype_jpg_quality, TRUE);
  jpeg_start_compress(&cinfo, TRUE);
  row_stride = cinfo.input_components * vp->video.width;
  while (cinfo.next_scanline < cinfo.image_height)
  {
    row_pointer[0] = &vp->pic[cinfo.next_scanline*row_stride];
    jpeg_write_scanlines(&cinfo,row_pointer,1);
  }
  jpeg_finish_compress(&cinfo);
  jpeg_destroy_compress(&cinfo);

  fclose(outfile);
  return;
}

/* ************************************************************** */
int video_dump_pict( struct VideoPaint *vp)
{
  fprintf(stdout,"Dumping...\n");
  switch (vp->option.savetype)
  {
  case JPEG : video_jpeg_save(vp); break;
  case PNG : video_png_save(vp); break;
  case PPM : video_ppm_save(vp); break;
  }
  return 0;
}
